@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2019 - 2020 Andri Yngvason
+ * Copyright (c) 2019 - 2022 Andri Yngvason
  *
  * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -43,15 +43,12 @@
 #include "screencopy.h"
 #include "data-control.h"
 #include "strlcpy.h"
-#include "logging.h"
 #include "output.h"
 #include "pointer.h"
 #include "keyboard.h"
 #include "seat.h"
 #include "cfg.h"
-#include "pixman-renderer.h"
 #include "transform-util.h"
-#include "damage-refinery.h"
 #include "usdt.h"
 
 #ifdef ENABLE_PAM
@@ -94,22 +91,18 @@ struct wayvnc {
 
 	struct nvnc* nvnc;
 	struct nvnc_display* nvnc_display;
-	struct nvnc_fb* buffer;
-
-	struct damage_refinery damage_refinery;
-	struct pixman_region16 current_damage;
 
 	const char* kb_layout;
 	const char* kb_variant;
 
 	uint32_t damage_area_sum;
 	uint32_t n_frames_captured;
-	uint32_t n_frames_rendered;
+
+	bool disable_input;
 };
 
 void wayvnc_exit(struct wayvnc* self);
 void on_capture_done(struct screencopy* sc);
-static void on_render(struct nvnc_display* display, struct nvnc_fb* fb);
 
 #if defined(GIT_VERSION)
 static const char wayvnc_version[] = GIT_VERSION;
@@ -123,6 +116,54 @@ struct wl_shm* wl_shm = NULL;
 struct zwp_linux_dmabuf_v1* zwp_linux_dmabuf = NULL;
 struct gbm_device* gbm_device = NULL;
 
+static bool registry_add_input(void* data, struct wl_registry* registry,
+			 uint32_t id, const char* interface,
+			 uint32_t version)
+{
+	struct wayvnc* self = data;
+
+	if (self->disable_input)
+		return false;
+
+	if (strcmp(interface, wl_seat_interface.name) == 0) {
+		struct wl_seat* wl_seat =
+			wl_registry_bind(registry, id, &wl_seat_interface, 7);
+		if (!wl_seat)
+			return true;
+
+		struct seat* seat = seat_new(wl_seat, id);
+		if (!seat) {
+			wl_seat_destroy(wl_seat);
+			return true;
+		}
+
+		wl_list_insert(&self->seats, &seat->link);
+		return true;
+	}
+
+	if (strcmp(interface, zwlr_virtual_pointer_manager_v1_interface.name) == 0) {
+		self->pointer_manager = wl_registry_bind(registry, id,
+				&zwlr_virtual_pointer_manager_v1_interface,
+				MIN(2, version));
+		return true;
+	}
+
+	if (strcmp(interface, zwp_virtual_keyboard_manager_v1_interface.name) == 0) {
+		self->keyboard_manager = wl_registry_bind(registry, id,
+				&zwp_virtual_keyboard_manager_v1_interface,
+				1);
+		return true;
+	}
+
+	if (strcmp(interface, zwlr_data_control_manager_v1_interface.name) == 0) {
+		self->data_control.manager = wl_registry_bind(registry, id,
+				&zwlr_data_control_manager_v1_interface, 2);
+		return true;
+	}
+
+	return false;
+}
+
 static void registry_add(void* data, struct wl_registry* registry,
 			 uint32_t id, const char* interface,
 			 uint32_t version)
@@ -150,11 +191,6 @@ static void registry_add(void* data, struct wl_registry* registry,
 		return;
 	}
 
-	if (strcmp(interface, wl_shm_interface.name) == 0) {
-		wl_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
-		return;
-	}
-
 	if (strcmp(interface, zwlr_screencopy_manager_v1_interface.name) == 0) {
 		self->screencopy.manager =
 			wl_registry_bind(registry, id,
@@ -163,35 +199,8 @@ static void registry_add(void* data, struct wl_registry* registry,
 		return;
 	}
 
-	if (strcmp(interface, zwlr_virtual_pointer_manager_v1_interface.name) == 0) {
-		self->pointer_manager =
-			wl_registry_bind(registry, id,
-					 &zwlr_virtual_pointer_manager_v1_interface,
-					 MIN(2, version));
-		return;
-	};
-
-	if (strcmp(interface, wl_seat_interface.name) == 0) {
-		struct wl_seat* wl_seat =
-			wl_registry_bind(registry, id, &wl_seat_interface, 7);
-		if (!wl_seat)
-			return;
-
-		struct seat* seat = seat_new(wl_seat, id);
-		if (!seat) {
-			wl_seat_destroy(wl_seat);
-			return;
-		}
-
-		wl_list_insert(&self->seats, &seat->link);
-		return;
-	}
-
-	if (strcmp(interface, zwp_virtual_keyboard_manager_v1_interface.name) == 0) {
-		self->keyboard_manager =
-			wl_registry_bind(registry, id,
-			                 &zwp_virtual_keyboard_manager_v1_interface,
-			                 1);
+	if (strcmp(interface, wl_shm_interface.name) == 0) {
+		wl_shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
 		return;
 	}
 
@@ -201,11 +210,8 @@ static void registry_add(void* data, struct wl_registry* registry,
 		return;
 	}
 
-	if (strcmp(interface, zwlr_data_control_manager_v1_interface.name) == 0) {
-		self->data_control.manager = wl_registry_bind(registry, id,
-				&zwlr_data_control_manager_v1_interface, 2);
+	if (registry_add_input(data, registry, id, interface, version))
 		return;
-	}
 }
 
 static void registry_remove(void* data, struct wl_registry* registry,
@@ -219,7 +225,7 @@ static void registry_remove(void* data, struct wl_registry* registry,
 		output_destroy(out);
 
 		if (out == self->selected_output) {
-			log_error("Selected output went away. Exiting...\n");
+			nvnc_log(NVNC_LOG_ERROR, "Selected output went away. Exiting...");
 			wayvnc_exit(self);
 		}
 
@@ -232,7 +238,7 @@ static void registry_remove(void* data, struct wl_registry* registry,
 		seat_destroy(seat);
 
 		if (seat == self->selected_seat) {
-			log_error("Selected seat went away. Exiting...\n");
+			nvnc_log(NVNC_LOG_ERROR, "Selected seat went away. Exiting...");
 			wayvnc_exit(self);
 		}
 
@@ -291,17 +297,26 @@ void wayvnc_destroy(struct wayvnc* self)
 
 	wl_shm_destroy(wl_shm);
 
-	zwp_virtual_keyboard_v1_destroy(self->keyboard_backend.virtual_keyboard);
-	zwp_virtual_keyboard_manager_v1_destroy(self->keyboard_manager);
-	keyboard_destroy(&self->keyboard_backend);
+	if (self->keyboard_backend.virtual_keyboard) {
+		zwp_virtual_keyboard_v1_destroy(
+				self->keyboard_backend.virtual_keyboard);
+		keyboard_destroy(&self->keyboard_backend);
+	}
 
-	zwlr_virtual_pointer_manager_v1_destroy(self->pointer_manager);
-	pointer_destroy(&self->pointer_backend);
+	if (self->keyboard_manager)
+		zwp_virtual_keyboard_manager_v1_destroy(self->keyboard_manager);
+
+	if (self->pointer_backend.pointer)
+		pointer_destroy(&self->pointer_backend);
+
+	if (self->pointer_manager)
+		zwlr_virtual_pointer_manager_v1_destroy(self->pointer_manager);
+
+	if (self->data_control.manager)
+		zwlr_data_control_manager_v1_destroy(self->data_control.manager);
 
 	if (self->screencopy.manager)
 		zwlr_screencopy_manager_v1_destroy(self->screencopy.manager);
-	if (self->data_control.manager)
-		zwlr_data_control_manager_v1_destroy(self->data_control.manager);
 
 	wl_registry_destroy(self->registry);
 	wl_display_disconnect(self->display);
@@ -344,13 +359,15 @@ static int init_wayland(struct wayvnc* self)
 
 	init_xdg_outputs(self);
 
-	if (!self->pointer_manager) {
-		log_error("Virtual Pointer protocol not supported by compositor.\n");
+	if (!self->pointer_manager && !self->disable_input) {
+		nvnc_log(NVNC_LOG_ERROR, "Virtual Pointer protocol not supported by compositor.");
+		nvnc_log(NVNC_LOG_ERROR, "wayvnc may still work if started with --disable-input.");
 		goto failure;
 	}
 
-	if (!self->keyboard_manager) {
-		log_error("Virtual Keyboard protocol not supported by compositor.\n");
+	if (!self->keyboard_manager && !self->disable_input) {
+		nvnc_log(NVNC_LOG_ERROR, "Virtual Keyboard protocol not supported by compositor.");
+		nvnc_log(NVNC_LOG_ERROR, "wayvnc may still work if started with --disable-input.");
 		goto failure;
 	}
 
@@ -358,7 +375,7 @@ static int init_wayland(struct wayvnc* self)
 	wl_display_roundtrip(self->display);
 
 	if (!self->screencopy.manager) {
-		log_error("Compositor doesn't support screencopy! Exiting.\n");
+		nvnc_log(NVNC_LOG_ERROR, "Compositor doesn't support screencopy! Exiting.");
 		goto failure;
 	}
 
@@ -381,15 +398,15 @@ void on_wayland_event(void* obj)
 
 	if (wl_display_read_events(self->display) < 0) {
 		if (errno == EPIPE || errno == ECONNRESET) {
-			log_error("Compositor has gone away. Exiting...\n");
+			nvnc_log(NVNC_LOG_ERROR, "Compositor has gone away. Exiting...");
 			wayvnc_exit(self);
 		} else {
-			log_error("Failed to read wayland events: %m\n");
+			nvnc_log(NVNC_LOG_ERROR, "Failed to read wayland events: %m");
 		}
 	}
 
 	if (wl_display_dispatch_pending(self->display) < 0)
-		log_error("Failed to dispatch pending\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to dispatch pending");
 }
 
 void wayvnc_exit(struct wayvnc* self)
@@ -467,7 +484,9 @@ static void on_client_cut_text(struct nvnc* server, const char* text, uint32_t l
 {
 	struct wayvnc* wayvnc = nvnc_get_userdata(server);
 
-	data_control_to_clipboard(&wayvnc->data_control, text, len);
+	if (!wayvnc->disable_input && wayvnc->data_control.manager) {
+		data_control_to_clipboard(&wayvnc->data_control, text, len);
+	}
 }
 
 bool on_auth(const char* username, const char* password, void* ud)
@@ -487,11 +506,25 @@ bool on_auth(const char* username, const char* password, void* ud)
 	return true;
 }
 
-int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port)
+static struct nvnc_fb* create_placeholder_buffer(uint16_t width, uint16_t height)
 {
-	self->nvnc = nvnc_open(addr, port);
+	uint16_t stride = width;
+	struct nvnc_fb* fb = nvnc_fb_new(width, height, DRM_FORMAT_XRGB8888,
+			stride);
+	if (!fb)
+		return NULL;
+
+	size_t size = nvnc_fb_get_pixel_size(fb) * height * stride;
+	memset(nvnc_fb_get_addr(fb), 0x60, size);
+
+	return fb;
+}
+
+int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port, bool is_unix)
+{
+	self->nvnc = is_unix ? nvnc_open_unix(addr) : nvnc_open(addr, port);
 	if (!self->nvnc) {
-		log_error("Failed to bind to address\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to bind to address");
 		return -1;
 	}
 
@@ -501,20 +534,18 @@ int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port)
 
 	nvnc_add_display(self->nvnc, self->nvnc_display);
 
-	nvnc_set_userdata(self->nvnc, self);
+	nvnc_set_userdata(self->nvnc, self, NULL);
 
 	nvnc_set_name(self->nvnc, "WayVNC");
 
-	nvnc_display_set_render_fn(self->nvnc_display, on_render);
-
 	if (self->cfg.enable_auth &&
 	    nvnc_enable_auth(self->nvnc, self->cfg.private_key_file,
 	                     self->cfg.certificate_file, on_auth, self) < 0) {
-		log_error("Failed to enable authentication\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to enable authentication");
 		goto failure;
 	}
 
-	if (self->pointer_manager)
+	if (self->pointer_backend.pointer)
 		nvnc_set_pointer_fn(self->nvnc, on_pointer_event);
 
 	if (self->keyboard_backend.virtual_keyboard) {
@@ -524,6 +555,23 @@ int init_nvnc(struct wayvnc* self, const char* addr, uint16_t port)
 
 	nvnc_set_cut_text_receive_fn(self->nvnc, on_client_cut_text);
 
+	struct nvnc_fb* placeholder_fb =
+		create_placeholder_buffer(self->selected_output->width,
+				self->selected_output->height);
+	if (!placeholder_fb) {
+		nvnc_log(NVNC_LOG_ERROR, "Failed to allocate a placeholder buffer");
+		goto failure;
+	}
+
+	struct pixman_region16 damage;
+	pixman_region_init_rect(&damage, 0, 0,
+			nvnc_fb_get_width(placeholder_fb),
+			nvnc_fb_get_height(placeholder_fb));
+
+	nvnc_display_feed_buffer(self->nvnc_display, placeholder_fb, &damage);
+	pixman_region_fini(&damage);
+	nvnc_fb_unref(placeholder_fb);
+
 	return 0;
 
 failure:
@@ -535,7 +583,7 @@ int wayvnc_start_capture(struct wayvnc* self)
 {
 	int rc = screencopy_start(&self->screencopy);
 	if (rc < 0) {
-		log_error("Failed to start capture. Exiting...\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to start capture. Exiting...");
 		wayvnc_exit(self);
 	}
 	return rc;
@@ -545,50 +593,19 @@ int wayvnc_start_capture_immediate(struct wayvnc* self)
 {
 	int rc = screencopy_start_immediate(&self->screencopy);
 	if (rc < 0) {
-		log_error("Failed to start capture. Exiting...\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to start capture. Exiting...");
 		wayvnc_exit(self);
 	}
 	return rc;
 }
 
-static void on_render(struct nvnc_display* display, struct nvnc_fb* fb)
-{
-	struct nvnc* nvnc = nvnc_display_get_server(display);
-	struct wayvnc* self = nvnc_get_userdata(nvnc);
-
-	if (!self->screencopy.back)
-		return;
-
-	self->n_frames_rendered++;
-
-	DTRACE_PROBE(wayvnc, render_start);
-
-	enum wl_output_transform transform = self->selected_output->transform;
-	wv_pixman_render(fb, self->screencopy.back, transform,
-			&self->current_damage);
-	pixman_region_clear(&self->current_damage);
-
-	DTRACE_PROBE(wayvnc, render_end);
-}
-
-static void wayvnc_damage_region(struct wayvnc* self,
-                                 struct pixman_region16* damage)
-{
-	pixman_region_union(&self->current_damage, &self->current_damage, damage);
-	nvnc_display_damage_region(self->nvnc_display, damage);
-}
-
 // TODO: Handle transform change too
 void on_output_dimension_change(struct output* output)
 {
 	struct wayvnc* self = output->userdata;
 	assert(self->selected_output == output);
 
-	// The buffer must have been set as a result of a screencopy frame
-	if (!self->buffer)
-		return;
-
-	log_debug("Output dimensions changed. Restarting frame capturer...\n");
+	nvnc_log(NVNC_LOG_DEBUG, "Output dimensions changed. Restarting frame capturer...");
 
 	screencopy_stop(&self->screencopy);
 	wayvnc_start_capture_immediate(self);
@@ -613,59 +630,38 @@ static uint32_t calculate_region_area(struct pixman_region16* region)
 
 void wayvnc_process_frame(struct wayvnc* self)
 {
-	uint32_t width = output_get_transformed_width(self->selected_output);
-	uint32_t height = output_get_transformed_height(self->selected_output);
-	uint32_t format = self->screencopy.back->format;
-
-	if ((int)self->selected_output->width != self->screencopy.back->width
-	   || (int)self->selected_output->height != self->screencopy.back->height) {
-		log_debug("Frame dimensions don't match output. Recapturing frame...\n");
-		wayvnc_start_capture_immediate(self);
-		return;
-	}
-
-	if (self->buffer && (width != nvnc_fb_get_width(self->buffer) ||
-			height != nvnc_fb_get_height(self->buffer))) {
-		log_debug("Frame dimensions changed. Resizing...\n");
-
-		nvnc_fb_unref(self->buffer);
-		self->buffer = NULL;
-		damage_refinery_destroy(&self->damage_refinery);
-	}
-
-	// TODO: Maybe the buffer needs to be changed inside on_render?
-	if (!self->buffer) {
-		self->buffer = nvnc_fb_new(width, height, format);
-		nvnc_display_set_buffer(self->nvnc_display, self->buffer);
-		log_debug("Set nvnc buffer: %p\n", self->buffer);
-
-		damage_refinery_init(&self->damage_refinery,
-			self->screencopy.back->width,
-			self->screencopy.back->height);
-	}
+	struct wv_buffer* buffer = self->screencopy.back;
+	self->screencopy.back = NULL;
 
 	self->n_frames_captured++;
 	self->damage_area_sum +=
-		calculate_region_area(&self->screencopy.back->damage);
+		calculate_region_area(&buffer->damage);
 
-	DTRACE_PROBE(wayvnc, refine_damage_start);
+	struct pixman_region16 damage;
+	pixman_region_init(&damage);
 
-	struct pixman_region16 txdamage, refined;
-	pixman_region_init(&txdamage);
-	pixman_region_init(&refined);
-	damage_refine(&self->damage_refinery, &refined,
-			&self->screencopy.back->damage,
-			self->screencopy.back);
-	wv_region_transform(&txdamage, &refined,
-			self->selected_output->transform,
-			self->selected_output->width,
-			self->selected_output->height);
-//	damage_dump(stdout, &txdamage, width, height, 32);
-	wayvnc_damage_region(self, &txdamage);
-	pixman_region_fini(&refined);
-	pixman_region_fini(&txdamage);
+	enum wl_output_transform output_transform, buffer_transform;
+	output_transform = self->selected_output->transform;
 
-	DTRACE_PROBE(wayvnc, refine_damage_end);
+	if (buffer->y_inverted) {
+		buffer_transform = wv_output_transform_compose(output_transform,
+				WL_OUTPUT_TRANSFORM_FLIPPED_180);
+
+		wv_region_transform(&damage, &buffer->damage,
+				WL_OUTPUT_TRANSFORM_FLIPPED_180,
+				buffer->width, buffer->height);
+	} else {
+		buffer_transform = output_transform;
+		pixman_region_copy(&damage, &buffer->damage);
+	}
+
+	nvnc_fb_set_transform(buffer->nvnc_fb,
+			(enum nvnc_transform)buffer_transform);
+
+	nvnc_display_feed_buffer(self->nvnc_display, buffer->nvnc_fb,
+			&damage);
+
+	pixman_region_fini(&damage);
 
 	wayvnc_start_capture(self);
 }
@@ -680,7 +676,7 @@ void on_capture_done(struct screencopy* sc)
 	case SCREENCOPY_IN_PROGRESS:
 		break;
 	case SCREENCOPY_FATAL:
-		log_error("Fatal error while capturing. Exiting...\n");
+		nvnc_log(NVNC_LOG_ERROR, "Fatal error while capturing. Exiting...");
 		wayvnc_exit(self);
 		break;
 	case SCREENCOPY_FAILED:
@@ -698,6 +694,7 @@ int wayvnc_usage(FILE* stream, int rc)
 "Usage: wayvnc [options] [address [port]]\n"
 "\n"
 "    -C,--config=<path>                        Select a config file.\n"
+"    -g,--gpu                                  Enable feature that need GPU.\n"
 "    -o,--output=<name>                        Select output to capture.\n"
 "    -k,--keyboard=<layout>[-<variant>]        Select keyboard layout with an\n"
 "                                              optional variant.\n"
@@ -705,7 +702,15 @@ int wayvnc_usage(FILE* stream, int rc)
 "    -r,--render-cursor                        Enable overlay cursor rendering.\n"
 "    -f,--max-fps=<fps>                        Set the rate limit (default 30).\n"
 "    -p,--show-performance                     Show performance counters.\n"
+"    -u,--unix-socket                          Create a UNIX domain socket\n"
+"                                              instead of TCP.\n"
+"    -d,--disable-input                        Disable all remote input\n"
 "    -V,--version                              Show version info.\n"
+"    -v,--verbose                              Be more verbose. Same as setting\n"
+"                                              --log-level=info.\n"
+"    -L,--log-level=<level>                    Set log level. The levels are:\n"
+"                                              error, warning, info, debug,\n"
+"                                              trace and quiet.\n"
 "    -h,--help                                 Get help (this text).\n"
 "\n";
 
@@ -720,26 +725,26 @@ int check_cfg_sanity(struct cfg* cfg)
 		int rc = 0;
 
 		if (!nvnc_has_auth()) {
-			log_error("Authentication can't be enabled because it was not selected during build\n");
+			nvnc_log(NVNC_LOG_ERROR, "Authentication can't be enabled because it was not selected during build");
 			return -1;
 		}
 
 		if (!cfg->certificate_file) {
-			log_error("Authentication enabled, but missing certificate_file\n");
+			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing certificate_file");
 			rc = -1;
 		}
 
 		if (!cfg->private_key_file) {
-			log_error("Authentication enabled, but missing private_key_file\n");
+			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing private_key_file");
 			rc = -1;
 		}
 		if (!cfg->username && !cfg->enable_pam) {
-			log_error("Authentication enabled, but missing username\n");
+			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing username");
 			rc = -1;
 		}
 
 		if (!cfg->password && !cfg->enable_pam) {
-			log_error("Authentication enabled, but missing password\n");
+			nvnc_log(NVNC_LOG_ERROR, "Authentication enabled, but missing password");
 			rc = -1;
 		}
 		return rc;
@@ -756,12 +761,10 @@ static void on_perf_tick(void* obj)
 	double area_avg = (double)self->damage_area_sum / (double)self->n_frames_captured;
 	double relative_area_avg = 100.0 * area_avg / total_area;
 
-	printf("Frames captured: %"PRIu32", rendered: %"PRIu32", average reported frame damage: %.1f %%\n",
-			self->n_frames_captured, self->n_frames_rendered,
-			relative_area_avg);
+	nvnc_log(NVNC_LOG_INFO, "Frames captured: %"PRIu32", average reported frame damage: %.1f %%",
+			self->n_frames_captured, relative_area_avg);
 
 	self->n_frames_captured = 0;
-	self->n_frames_rendered = 0;
 	self->damage_area_sum = 0;
 }
 
@@ -788,6 +791,17 @@ void parse_keyboard_option(struct wayvnc* self, char* arg)
 	self->kb_layout = arg;
 }
 
+static int log_level_from_string(const char* str)
+{
+	if (0 == strcmp(str, "quiet")) return NVNC_LOG_PANIC;
+	if (0 == strcmp(str, "error")) return NVNC_LOG_ERROR;
+	if (0 == strcmp(str, "warning")) return NVNC_LOG_WARNING;
+	if (0 == strcmp(str, "info")) return NVNC_LOG_INFO;
+	if (0 == strcmp(str, "debug")) return NVNC_LOG_DEBUG;
+	if (0 == strcmp(str, "trace")) return NVNC_LOG_TRACE;
+	return -1;
+}
+
 int show_version(void)
 {
 	printf("wayvnc: %s\n", wayvnc_version);
@@ -801,9 +815,11 @@ int main(int argc, char* argv[])
 	struct wayvnc self = { 0 };
 
 	const char* cfg_file = NULL;
+	bool enable_gpu_features = false;
 
 	const char* address = NULL;
 	int port = 0;
+	bool use_unix_socket = false;
 
 	const char* output_name = NULL;
 	const char* seat_name = NULL;
@@ -811,12 +827,16 @@ int main(int argc, char* argv[])
 	bool overlay_cursor = false;
 	bool show_performance = false;
 	int max_rate = 30;
+	bool disable_input = false;
 
-	static const char* shortopts = "C:o:k:s:rf:hpV";
+	static const char* shortopts = "C:go:k:s:rf:hpudVvL:";
 	int drm_fd MAYBE_UNUSED = -1;
 
+	int log_level = NVNC_LOG_WARNING;
+
 	static const struct option longopts[] = {
 		{ "config", required_argument, NULL, 'C' },
+		{ "gpu", no_argument, NULL, 'g' },
 		{ "output", required_argument, NULL, 'o' },
 		{ "keyboard", required_argument, NULL, 'k' },
 		{ "seat", required_argument, NULL, 's' },
@@ -824,7 +844,11 @@ int main(int argc, char* argv[])
 		{ "max-fps", required_argument, NULL, 'f' },
 		{ "help", no_argument, NULL, 'h' },
 		{ "show-performance", no_argument, NULL, 'p' },
+		{ "unix-socket", no_argument, NULL, 'u' },
+		{ "disable-input", no_argument, NULL, 'd' },
 		{ "version", no_argument, NULL, 'V' },
+		{ "verbose", no_argument, NULL, 'v' },
+		{ "log-level", required_argument, NULL, 'L' },
 		{ NULL, 0, NULL, 0 }
 	};
 
@@ -837,6 +861,9 @@ int main(int argc, char* argv[])
 		case 'C':
 			cfg_file = optarg;
 			break;
+		case 'g':
+			enable_gpu_features = true;
+			break;
 		case 'o':
 			output_name = optarg;
 			break;
@@ -855,6 +882,23 @@ int main(int argc, char* argv[])
 		case 'p':
 			show_performance = true;
 			break;
+		case 'u':
+			use_unix_socket = true;
+			break;
+		case 'd':
+			disable_input = true;
+			break;
+		case 'v':
+			log_level = NVNC_LOG_INFO;
+			break;
+		case 'L':
+			log_level = log_level_from_string(optarg);
+			if (log_level < 0) {
+				fprintf(stderr, "Invalid log level: %s\n",
+						optarg);
+				return wayvnc_usage(stderr, 1);
+			}
+			break;
 		case 'V':
 			return show_version();
 		case 'h':
@@ -864,6 +908,8 @@ int main(int argc, char* argv[])
 		}
 	}
 
+	nvnc_set_log_level(log_level);
+
 	int n_args = argc - optind;
 
 	if (n_args >= 1)
@@ -872,14 +918,19 @@ int main(int argc, char* argv[])
 	if (n_args >= 2)
 		port = atoi(argv[optind + 1]);
 
+	if (seat_name && disable_input) {
+		nvnc_log(NVNC_LOG_ERROR, "seat and disable-input are conflicting options");
+		return 1;
+	}
+
 	errno = 0;
 	int cfg_rc = cfg_load(&self.cfg, cfg_file);
 	if (cfg_rc != 0 && (cfg_file || errno != ENOENT)) {
 		if (cfg_rc > 0) {
-			log_error("Failed to load config. Error on line %d\n",
+			nvnc_log(NVNC_LOG_ERROR, "Failed to load config. Error on line %d",
 			          cfg_rc);
 		} else {
-			log_error("Failed to load config. %m\n");
+			nvnc_log(NVNC_LOG_ERROR, "Failed to load config. %m");
 		}
 
 		return 1;
@@ -896,8 +947,10 @@ int main(int argc, char* argv[])
 	if (!address) address = DEFAULT_ADDRESS;
 	if (!port) port = DEFAULT_PORT;
 
+	self.disable_input = disable_input;
+
 	if (init_wayland(&self) < 0) {
-		log_error("Failed to initialise wayland\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise wayland");
 		return 1;
 	}
 
@@ -905,28 +958,28 @@ int main(int argc, char* argv[])
 	if (output_name) {
 		out = output_find_by_name(&self.outputs, output_name);
 		if (!out) {
-			log_error("No such output\n");
+			nvnc_log(NVNC_LOG_ERROR, "No such output");
 			goto failure;
 		}
 	} else {
 		out = output_first(&self.outputs);
 		if (!out) {
-			log_error("No output found\n");
+			nvnc_log(NVNC_LOG_ERROR, "No output found");
 			goto failure;
 		}
 	}
 
-	struct seat* seat;
+	struct seat* seat = NULL;
 	if (seat_name) {
 		seat = seat_find_by_name(&self.seats, seat_name);
 		if (!seat) {
-			log_error("No such seat\n");
+			nvnc_log(NVNC_LOG_ERROR, "No such seat");
 			goto failure;
 		}
-	} else {
+	} else if (!self.disable_input) {
 		seat = seat_first(&self.seats);
 		if (!seat) {
-			log_error("No seat found\n");
+			nvnc_log(NVNC_LOG_ERROR, "No seat found");
 			goto failure;
 		}
 	}
@@ -935,43 +988,50 @@ int main(int argc, char* argv[])
 	self.selected_seat = seat;
 	self.screencopy.wl_output = out->wl_output;
 	self.screencopy.rate_limit = max_rate;
+	self.screencopy.enable_linux_dmabuf = enable_gpu_features;
 
-	self.keyboard_backend.virtual_keyboard =
-		zwp_virtual_keyboard_manager_v1_create_virtual_keyboard(
-			self.keyboard_manager, self.selected_seat->wl_seat);
+	if (self.keyboard_manager) {
+		self.keyboard_backend.virtual_keyboard =
+			zwp_virtual_keyboard_manager_v1_create_virtual_keyboard(
+				self.keyboard_manager, self.selected_seat->wl_seat);
 
-	struct xkb_rule_names rule_names = {
-		.rules = self.cfg.xkb_rules,
-		.layout = self.kb_layout ? self.kb_layout : self.cfg.xkb_layout,
-		.model = self.cfg.xkb_model ? self.cfg.xkb_model : "pc105",
-		.variant = self.kb_variant ? self.kb_variant :
-			self.cfg.xkb_variant,
-		.options = self.cfg.xkb_options,
-	};
+		struct xkb_rule_names rule_names = {
+			.rules = self.cfg.xkb_rules,
+			.layout = self.kb_layout ? self.kb_layout : self.cfg.xkb_layout,
+			.model = self.cfg.xkb_model ? self.cfg.xkb_model : "pc105",
+			.variant = self.kb_variant ? self.kb_variant :
+				self.cfg.xkb_variant,
+			.options = self.cfg.xkb_options,
+		};
 
-	keyboard_init(&self.keyboard_backend, &rule_names);
+		if (keyboard_init(&self.keyboard_backend, &rule_names) < 0) {
+			nvnc_log(NVNC_LOG_ERROR, "Failed to initialise keyboard");
+			goto failure;
+		}
+	}
 
 	self.pointer_backend.vnc = self.nvnc;
 	self.pointer_backend.output = self.selected_output;
 
-	int pointer_manager_version =
-		zwlr_virtual_pointer_manager_v1_get_version(self.pointer_manager);
+	if (self.pointer_manager) {
+		int pointer_manager_version =
+			zwlr_virtual_pointer_manager_v1_get_version(self.pointer_manager);
 
-	self.pointer_backend.pointer = pointer_manager_version >= 2
-		? zwlr_virtual_pointer_manager_v1_create_virtual_pointer_with_output(
-			self.pointer_manager, self.selected_seat->wl_seat,
-			out->wl_output)
-		: zwlr_virtual_pointer_manager_v1_create_virtual_pointer(
-			self.pointer_manager, self.selected_seat->wl_seat);
-
-	pointer_init(&self.pointer_backend);
+		self.pointer_backend.pointer = pointer_manager_version >= 2
+			? zwlr_virtual_pointer_manager_v1_create_virtual_pointer_with_output(
+				self.pointer_manager, self.selected_seat->wl_seat,
+				out->wl_output)
+			: zwlr_virtual_pointer_manager_v1_create_virtual_pointer(
+				self.pointer_manager, self.selected_seat->wl_seat);
 
+		pointer_init(&self.pointer_backend);
+	}
 	out->on_dimension_change = on_output_dimension_change;
 	out->userdata = &self;
 
 #ifdef ENABLE_SCREENCOPY_DMABUF
 	if (init_render_node(&drm_fd) < 0) {
-		log_error("Failed to initialise DRM render node. No GPU acceleration will be available.\n");
+		nvnc_log(NVNC_LOG_ERROR, "Failed to initialise DRM render node. No GPU acceleration will be available.");
 	}
 #endif
 
@@ -984,14 +1044,14 @@ int main(int argc, char* argv[])
 	if (init_main_loop(&self) < 0)
 		goto main_loop_failure;
 
-	if (init_nvnc(&self, address, port) < 0)
+	if (init_nvnc(&self, address, port, use_unix_socket) < 0)
 		goto nvnc_failure;
 
 	if (self.screencopy.manager)
 		screencopy_init(&self.screencopy);
 
 	if (!self.screencopy.manager) {
-		log_error("screencopy is not supported by compositor\n");
+		nvnc_log(NVNC_LOG_ERROR, "screencopy is not supported by compositor");
 		goto capture_failure;
 	}
 
@@ -999,8 +1059,6 @@ int main(int argc, char* argv[])
 		data_control_init(&self.data_control, self.display, self.nvnc,
 				self.selected_seat->wl_seat);
 
-	pixman_region_init(&self.current_damage);
-
 	self.screencopy.overlay_cursor = overlay_cursor;
 
 	if (wayvnc_start_capture(&self) < 0)
@@ -1019,13 +1077,6 @@ int main(int argc, char* argv[])
 
 	screencopy_stop(&self.screencopy);
 
-	if (self.buffer) {
-		damage_refinery_destroy(&self.damage_refinery);
-		nvnc_fb_unref(self.buffer);
-	}
-
-	pixman_region_fini(&self.current_damage);
-
 	nvnc_display_unref(self.nvnc_display);
 	nvnc_close(self.nvnc);
 	if (zwp_linux_dmabuf)
